/*
 * krt_netlink2.c
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>
 *
 */

/*
 * The interface to Linux netlink socket.
 */

#define	INCLUDE_IOCTL
#define	INCLUDE_IF
#include "include.h"
#ifdef	PROTO_INET
#include "inet.h"
#endif	/* PROTO_INET */
#include "krt.h"
#include "krt_var.h"

#include <sys/socket.h>
#include <linux/rtnetlink.h>
#include <linux/inetdevice.h>

#if	IFNAMSIZ > IFL_NAMELEN
#error - IFL_NAMELEN not compatible with IFNAMESIZ
#endif

/* Class in which this instanse of gated works.
   0 means RT_TABLE_MAIN+RT_TABLE_DEFAULT
 */

int class_1;

adv_entry *krt_import_list = NULL;
adv_entry *krt_export_list = NULL;


/* Protocol ID: just an arbitrary number */
#ifndef RTPROT_GATED
#define RTPROT_GATED 8
#endif

/* We are in rtread_state */
int rtnl_rtread_active;
struct sockaddr_nl rtnl_addr;
__u32 rtnl_seq = 1;
struct ucred *peer_creds;

int parse_rtattr(struct rtattr *tb[], int maxattr, struct rtattr *rta, int len)
{
	while (RTA_OK(rta, len)) {
		if (rta->rta_type <= maxattr)
			tb[rta->rta_type] = rta;
		rta = RTA_NEXT(rta,len);
	}
	return 0;
}

static int parse_nexthop(task *tp,
			 struct rtmsg *rtm,
			 rt_parms *rtp,
			 int i,
			 int ifindex,
			 struct rtattr **rtas,
			 struct rtattr **rtas2)
{
	sockaddr_un *lcl = NULL;
	flag_t s = rtp->rtp_state;
	if_addr *ifap;

	if (rtas2[RTA_GATEWAY]) {
		__u32 gw = *(__u32*)RTA_DATA(rtas2[RTA_GATEWAY]);
		rtp->rtp_routers[i] = sockbuild_in(0, gw);
		BIT_SET(s, RTS_EXTERIOR);
	} else if (rtas[RTA_PREFSRC]) {
		__u32 gw = *(__u32*)RTA_DATA(rtas[RTA_PREFSRC]);
		lcl = rtp->rtp_routers[i] = sockbuild_in(0, gw);
		BIT_SET(s, RTS_INTERIOR);
	}

	ifap = if_withindex_and_addrs(ifindex, lcl, rtp->rtp_routers[i]);
	if (!ifap) {
		if (rtp->rtp_routers[i] && (rtm->rtm_flags&RTNH_F_ONLINK))
			return -1;
		trace_log_tp(tp,
			     TRC_NL_BEFORE|TRC_NL_AFTER,
			     LOG_ERR,
			     ("krt_recv_route: requested route to %A/%d via off-link gateway; dev %d",
			      rtp->rtp_dest,
			      rtp->rtp_dest_mask_log,
			      ifindex));
		return -1;
	}

	if (rtp->rtp_routers[i] == NULL) {
		BIT_SET(s, RTS_INTERIOR);
		rtp->rtp_routers[i] = ifap->ifa_addr_local;
	} else if (rtp->rtp_routers[i] == ifap->ifa_addr_local) {
		BIT_RESET(s, RTS_EXTERIOR);
		BIT_SET(s, RTS_INTERIOR);
	} else if (rtas2[RTA_GATEWAY] && rtm->rtm_scope < RT_SCOPE_LINK) {
		BIT_SET(s, RTS_GATEWAY);
	}

	rtp->rtp_state |= s;

#ifdef KRT_REALMS
	if (rtp->rtp_realm == 0) {
		__u32 realm = 0;

		if (rtas[RTA_FLOW])
			realm = *(__u32*)RTA_DATA(rtas[RTA_FLOW]);
		else if (rtas2[RTA_FLOW])
			realm = *(__u32*)RTA_DATA(rtas2[RTA_FLOW]);

		if (realm < 256)
			rtp->rtp_realm = realm;
	} 
#endif

	return 0;
}


static int
parse_rtmsg(task *tp,
	    struct nlmsghdr *nlh, struct rtmsg *rtm,
	    rt_parms *rtp)
{
	struct rtattr *rtas[RTA_MAX+1];
	__u32 prefix;

	memset(rtp, 0, sizeof (*rtp));
	rtp->rtp_n_gw = 1;

	/* Echo of our own request; ignore it */

	if (nlh->nlmsg_pid == rtnl_addr.nl_pid)
		return -1;

	/* Ignore kernel interface routes: gated has its own (and wrong)
	   understanding of them, that it is difficult to override.
	 */

	if (rtm->rtm_protocol == RTPROT_KERNEL)
		return -1;

	/* Should be handled with "import kernel" */
	if (class_1 == 0) {
		/* Ignore all policy classes: gated does not understand them */
		if (rtm->rtm_table != RT_TABLE_MAIN &&
		    rtm->rtm_table != RT_TABLE_DEFAULT)
			return -1;
	} else {
		if (rtm->rtm_table != class_1)
			return -1;
	}

	/* Ignore TOS routes: gated does not understand them */
	if (rtm->rtm_tos)
		return -1;

	memset(rtas, 0, sizeof(rtas));
	parse_rtattr(rtas, RTA_MAX, RTM_RTA(rtm), RTM_PAYLOAD(nlh));

	rtp->rtp_dest_mask_log = rtm->rtm_dst_len;
	prefix = 0;
	if (rtas[RTA_DST])
		prefix = *(__u32*)RTA_DATA(rtas[RTA_DST]);
	rtp->rtp_dest = sockbuild_in(0, prefix);

	rtp->rtp_state = 0;

	if (rtm->rtm_protocol == RTPROT_STATIC) {
		rtp->rtp_gwp = krt_gwp;
		rtp->rtp_preference = RTPREF_KERNEL;
		BIT_SET(rtp->rtp_state, RTS_STATIC|RTS_RETAIN);
	} else if (rtm->rtm_protocol < RTPROT_STATIC ||
		   rtm->rtm_protocol == RTPROT_RA ||
		   rtm->rtm_protocol == RTPROT_GATED ||
		   rtm->rtm_protocol > 128) {
		rtp->rtp_gwp = krt_gwp_remnant;
		rtp->rtp_preference = RTPREF_KERNEL_REMNANT;
		BIT_SET(rtp->rtp_state, RTS_NOADVISE);
	} else {
		/* All others. We suppose, that it is another
		   routing daemon instance.
		   Mapping should be done in configuration file,
		   but for now we just assume the following:
		 */
		BIT_SET(rtp->rtp_state, RTS_RETAIN);
		rtp->rtp_gwp = krt_gwp;
#ifndef PROTO_OSPF
		rtp->rtp_preference = RTPREF_OSPF;
#else
		rtp->rtp_preference = RTPREF_OSPF+1;
#endif
	}

	if (rtm->rtm_type == RTN_UNREACHABLE) {
		BIT_SET(rtp->rtp_state, RTS_REJECT);
		rtp->rtp_router = inet_addr_loopback;
	} else if (rtm->rtm_type == RTN_BLACKHOLE) {
		BIT_SET(rtp->rtp_state, RTS_BLACKHOLE);
		rtp->rtp_router = inet_addr_loopback;
	} else if (rtm->rtm_type != RTN_UNICAST)
		return -1;

	if (rtas[RTA_OIF] == NULL) {
		trace_log_tp(tp,
			     TRC_NL_BEFORE|TRC_NL_AFTER,
			     LOG_ERR,
			     ("krt_recv_route: oif=NULL"));
		return -1;
	}

	if (rtas[RTA_MULTIPATH]) {
		int i = 0;
		struct rtnexthop *nh = RTA_DATA(rtas[RTA_MULTIPATH]);
		int mlen = RTA_PAYLOAD(rtas[RTA_MULTIPATH]);
		struct rtattr *rtas2[RTA_MAX+1];

		while (nh->rtnh_len >= sizeof(*nh) && nh->rtnh_len <= mlen) {
			int ifindex = nh->rtnh_ifindex;

			if (i >= RT_N_MULTIPATH)
				return -1;
			memset(rtas2, 0, sizeof(rtas2));
			parse_rtattr(rtas2, RTA_MAX, RTM_RTA(rtm), RTM_PAYLOAD(nlh));
			if (parse_nexthop(tp, rtm, rtp, i, ifindex, rtas, rtas2))
				return -1;
			i++;
			mlen -= RTA_ALIGN(nh->rtnh_len);
			nh = RTNH_NEXT(nh);
		}
		rtp->rtp_n_gw = i;
	} else {
		int ifindex = *(int*)RTA_DATA(rtas[RTA_OIF]);

		if (parse_nexthop(tp, rtm, rtp, 0, ifindex, rtas, rtas))
			return -1;
	}
	return 0;
}

static void
__krt_recv_route(task * tp, struct nlmsghdr *nlh, int err)
{
	struct rtmsg *rtm = NLMSG_DATA(nlh);
	char *msg = rtnl_rtread_active ? "RTREAD" : "RECV";
	char *errmsg = NULL;
	char *type = "ADD";

	int need_sync;
	rt_parms rtparms;

	if (parse_rtmsg(tp, nlh, rtm, &rtparms))
		return;

	if (!rtnl_rtread_active)
		rt_open(tp);

	need_sync = TRUE;

	switch (nlh->nlmsg_type) {
		rt_entry *rt_exists, *rt;
	case RTM_NEWROUTE:
		/* Attempt to find an existing route */
		rt = rt_locate(RTS_NETROUTE,
			       rtparms.rtp_dest,
			       rtparms.rtp_dest_mask_log,
			       RTPROTO_BIT_ANY);
		rt_exists = rt;
		if (rt) {
			/* We have a route to this destination */
			type = "CHANGE";

			/* Sync to the kernel */
			krt_rth_reset(rt->rt_head,
				      rtparms.rtp_state,
				      rtparms.rtp_n_gw,
				      rtparms.rtp_routers,
				      (if_addr **) 0
#ifdef KRT_REALMS
				      , rtparms.rtp_realm
#endif
				      );
			need_sync = FALSE;

			if (rt->rt_gwp->gw_proto != RTPROTO_KERNEL) {
				/* The best route is owned by gated.
				   Maybe, gated has some static kernel route however.
				 */
				rt = rt_locate(RTS_NETROUTE,
					       rtparms.rtp_dest,
					       rtparms.rtp_dest_mask_log,
					       RTPROTO_BIT(RTPROTO_KERNEL));

				if (rt == NULL) {
					/* No, it does not.
					   If new route is not remnant, add it to RIB.
					   Otherwise, ignore it, it will be deleted
					   from kernel FIB.
					 */
					if (rtparms.rtp_gwp == krt_gwp)
						goto Add;
					break;
				}
			}

			/* It is HACK to force gated to ignore multiple
			   routes at least, when synchronizing RIB.
			 */
			if (rtnl_rtread_active) {
				type = "APPEND";
				break;
			}

			/* Now rt is RTPROTO_KERNEL route */
			if (rtparms.rtp_gwp == rt->rt_gwp) {
				BIT_RESET(rt->rt_state, RTS_REJECT|RTS_BLACKHOLE|RTS_STATIC|RTS_RETAIN|RTS_NOADVISE);
				rt->rt_state |= BIT_TEST(rtparms.rtp_state, RTS_REJECT|RTS_BLACKHOLE|RTS_STATIC|RTS_RETAIN);
				if (rt->rt_gwp == krt_gwp_remnant) {
					BIT_SET(rt->rt_state, RTS_NOADVISE);
				}
				rt_change(rt,
					  (metric_t) 0,
					  (metric_t) 0,
					  (metric_t) 0,
					  rtparms.rtp_preference,
					  (pref_t) 0,
					  1,
					  &rtparms.rtp_router);
			} else {
				/* If new route is remnant, but we have some
				   static route, ignore remnant
				 */
				if (rtparms.rtp_gwp == krt_gwp_remnant)
					break;

				/* We have both static and remnant!!!
				   Actually, it is impossible. However,
				   we delete remnant and install static instead.
				 */
				rt_delete(rt);

				goto Add;
			}

			break;
		}

Add:
		rt = rt_add(&rtparms);
		if (!rt) {
			/* The add failed */
			errmsg = "failed to add";

			if (need_sync) {
				/* Delete the bogus entry */
				errmsg = "failed to add; deleting";

				if (rt_exists) {
					krt_rth_reset(rt_exists->rt_head,
						      rtparms.rtp_state,
						      rtparms.rtp_n_gw,
						      rtparms.rtp_routers,
						      (if_addr **) 0
#ifdef KRT_REALMS
						      , rtparms.rtp_realm
#endif
						      );

					need_sync = FALSE;
				} else {
					if (rtnl_rtread_active)
						krt_delq_add(&rtparms);
					else
						krt_delete_dst(tp,
							       rtparms.rtp_dest,
							       rtparms.rtp_dest_mask_log,
							       RTPROTO_KERNEL,
							       rtparms.rtp_state,
							       rtparms.rtp_n_gw,
							       rtparms.rtp_routers,
							       (if_addr **) 0);
				}
			}
		} else {
			/* The add succeeded */

			if (rtparms.rtp_gwp == krt_gwp_remnant) {
				/* Make sure the remnant timer is running */

				krt_age_create();
			}

			if (need_sync) {
				/* Indicate current state of kernel */
		
				krt_rth_reset(rt->rt_head,
					      rtparms.rtp_state,
					      rtparms.rtp_n_gw,
					      rtparms.rtp_routers,
					      (if_addr **) 0
#ifdef KRT_REALMS
					      , rtparms.rtp_realm
#endif
					      );

				need_sync = FALSE;
			}
			if (rt != rt->rt_active) {
				errmsg = "not the best one";

				if (!(rtparms.rtp_state&RTS_STATIC)) {

					rt_delete(rt);

					errmsg = "not the best one; deleting";
				}
			}
		}
		break;

	case RTM_DELROUTE:
		type = "DEL";
		if (err) {
			type = "ADDERR";
		}

		rt = rt_locate(RTS_NETROUTE,
			       rtparms.rtp_dest,
			       rtparms.rtp_dest_mask_log,
			       RTPROTO_BIT_ANY);
		if (rt) {
			krt_rth_reset(rt->rt_head,
				      (flag_t) 0,
				      0,
				      (sockaddr_un **) 0,
				      (if_addr **) 0
#ifdef KRT_REALMS
				      , NOREALM
#endif
				      );

			/* Remnant route deletions are always advertised 8)

			   Static routes are deleted from RIB only if they
			   were static in kernel.
			 */

			if ((rt->rt_gwp == krt_gwp &&
			     (!BIT_TEST(rt->rt_state, RTS_STATIC) ||
			      BIT_TEST(rtparms.rtp_state, RTS_STATIC))) ||
			    rt->rt_gwp == krt_gwp_remnant) {
				rt_delete(rt);
			} else {
				errmsg = "overriding attempt to delete";
			}
		} else {
			errmsg = "deleting unknown route";
		}
		break;
	}

	if (errmsg || TRACE_TP(tp, TR_KRT_REMNANTS)) {
		krt_trace(tp,
			  msg,
			  type,
			  rtparms.rtp_dest,
			  rtparms.rtp_dest_mask_log,
			  rtparms.rtp_router,
			  (flag_t) krt_state_to_flags(rtparms.rtp_state),
#ifdef KRT_REALMS
			  rtparms.rtp_realm,
#endif
			  errmsg,
			  errmsg ? LOG_ERR : LOG_DEBUG);
	}

	if (!rtnl_rtread_active)
		rt_close(tp, (gw_entry *) 0, 1, NULL);
}

static void
__krt_recv_ifaddr(task * tp, struct nlmsghdr *nlh, int err)
{
	struct ifaddrmsg *ifa = NLMSG_DATA(nlh);
	if_info ifi;
	if_link *ifl = ifl_locate_index(ifa->ifa_index);
	struct rtattr *rtas[IFA_MAX+1];
	int len = nlh->nlmsg_len;
	struct rtattr *rta = IFA_RTA(ifa);
	int brd_ok = 0;

	if (ifl == NULL) {
		trace_log_tp(tp,
			     TRC_NL_BEFORE|TRC_NL_AFTER,
			     LOG_ERR,
			     ("krt_recv_ifaddr: unknown link %d\n",
			      ifa->ifa_index));
		return;
	}

	if (ifa->ifa_family != AF_INET)
		return;

	bzero((caddr_t) &ifi, sizeof (ifi));

	len -= NLMSG_LENGTH(sizeof(*ifa));

	memset(rtas, 0, sizeof(rtas));
	parse_rtattr(rtas, IFA_MAX, rta, len);

	ifi.ifi_link = ifl;
	ifi.ifi_state = ifl->ifl_state;
	ifi.ifi_metric = ifl->ifl_metric;
	ifi.ifi_mtu = ifl->ifl_mtu;
	ifi.ifi_mask_log = ifa->ifa_prefixlen;
	ifi.ifi_netmask = inet_static_mask(ifi.ifi_mask_log);

	if (rtas[IFA_ADDRESS]) {
		__u32 data = *(__u32*)RTA_DATA(rtas[IFA_ADDRESS]);
		ifi.ifi_addr = ifi.ifi_addr_local = sockdup(sockbuild_in(0, data));
	}
	if (rtas[IFA_LOCAL]) {
		__u32 data = *(__u32*)RTA_DATA(rtas[IFA_LOCAL]);
		if (ifi.ifi_addr == NULL || sock2ip(ifi.ifi_addr) != data)
			ifi.ifi_addr_local = sockdup(sockbuild_in(0, data));
	}
	if (ifi.ifi_addr == NULL)
		ifi.ifi_addr = ifi.ifi_addr_local;
	if (ifi.ifi_addr_local == NULL) {
		/* Impossible event: read kernel bug */
		return;
	}

	memcpy(ifi.ifi_label, ifl->ifl_name, IFL_NAMELEN+1);
	if (rtas[IFA_LABEL])
		strcpy(ifi.ifi_label, RTA_DATA(rtas[IFA_LABEL]));
	if (strcmp(ifi.ifi_label, ifl->ifl_name))
		ifi.ifi_state |= IFS_SECONDARY;
	if (ifa->ifa_flags&IFA_F_SECONDARY)
		ifi.ifi_state |= IFS_SECONDARY;

	/* Gated has some bizarre internal understanding of PtP interfaces,
	   now we try to fake it to force interpret almost
	   all the combinations correctly.

	   Namely, we must to reduce the things to two cases:

	   !IFF_POINTOPOINT && (addr == addr_local)
	   or
	   IFF_POINTOPOINT

	   We should take into account that gated considers netmask
	   on IFF_POINTOPOINT interfaces only as a helper for RIP.
	 */

	if (ifi.ifi_addr_local == ifi.ifi_addr) {
		if (ifi.ifi_state&IFS_POINTOPOINT) {
			/* ifconfig ppp0 193.193.193.193 netmask 255.255.255.x */

			if (ifi.ifi_mask_log != 32) {
				/* F.e. ifconfig ppp0 193.193.193.193 netmask 255.255.255.0 */
				ifi.ifi_state &= ~IFS_POINTOPOINT;
				brd_ok = 1;
			} /* else:
			       ifconfig ppp0 193.193.193.193
			     i.e. the interface is looped back.
			   */
		}
	} else if (!(ifi.ifi_state&IFS_POINTOPOINT)) {
		if (ifi.ifi_mask_log == 32) {
			/* ifconfig xxx0 193.193.193.193 \
			   netmask 255.255.255.255 \
			   dstaddr 193.193.193.194
			*/
			ifi.ifi_state |= IFS_POINTOPOINT;
		} else {
			__u32 remote = sock2ip(ifi.ifi_addr);
			__u32 local = sock2ip(ifi.ifi_addr_local);

			if (remote == (remote&sock2ip(ifi.ifi_netmask))) {
				if ((local^remote)&sock2ip(ifi.ifi_netmask)) {
					trace_log_tp(tp,
						     TRC_NL_BEFORE|TRC_NL_AFTER,
						     LOG_ERR,
						     ("krt_recv_if: BAD ADDRESSES %s %A/%d local %A dev %s",
						      nlh->nlmsg_type == RTM_NEWADDR ? "add" : "del",
						      ifi.ifi_addr, ifi.ifi_mask_log,
						      ifi.ifi_addr_local, ifl->ifl_name));
					return;
				}
				sockfree(ifi.ifi_addr);
				ifi.ifi_addr = ifi.ifi_addr_local;
			} else {
				ifi.ifi_state |= IFS_POINTOPOINT;
			}
		}
	}

	if (ifi.ifi_state&IFS_POINTOPOINT)
		ifi.ifi_state &= ~IFS_BROADCAST;

	/* Uhh... It was weirdly complicated task :-) */

	if (ifi.ifi_state&IFS_BROADCAST || brd_ok) {
		if (rtas[IFA_BROADCAST]) {
			__u32 data = *(__u32*)RTA_DATA(rtas[IFA_BROADCAST]);
			ifi.ifi_addr_broadcast = sockdup(sockbuild_in(0, data));
			ifi.ifi_state |= IFS_BROADCAST;
		} else if (ifi.ifi_state&IFS_BROADCAST) {
			__u32 data = INADDR_BROADCAST;
			ifi.ifi_addr_broadcast = sockdup(sockbuild_in(0, data));
		}
	}

	trace_only_tp(tp,
		      0,
		      ("krt_recv_if: %s %A/%d local %A dev %s",
		       nlh->nlmsg_type == RTM_NEWADDR ? "add" : "del",
		       ifi.ifi_addr, ifi.ifi_mask_log,
		       ifi.ifi_addr_local, ifl->ifl_name));

	if (nlh->nlmsg_type == RTM_NEWADDR)
		if_conf_addaddr(tp, &ifi);
	else
		if_conf_deladdr(tp, &ifi);
}


flag_t krt_rt_support = KRTS_HOST|KRTS_VAR_MASK ;

static char rtnl_buffer[4096];
static int  rtnl_ptr;

static __inline__ void RTA_PUT32(struct nlmsghdr *nlh, int type, __u32 data)
{
	struct rtattr *rta = (struct rtattr*)((char*)nlh + NLMSG_ALIGN(nlh->nlmsg_len));

	rta->rta_len = RTA_LENGTH(4);
	rta->rta_type = type;
	*(__u32*)RTA_DATA(rta) = data;
	nlh->nlmsg_len += NLMSG_ALIGN(rta->rta_len);
}

static void rtnl_flush_buffer(task *tp)
{
	int status;
	struct sockaddr_nl nladdr;
	static struct iovec iov = { &rtnl_buffer, 0 };
	struct msghdr msg = {
		(void*)&nladdr, sizeof(nladdr),
		&iov,	1,
		NULL,	0,
		0
	};

	if (BIT_TEST(task_state, TASKS_TEST)
	     || BIT_TEST(krt_options, KRT_OPT_NOINSTALL)) {
		rtnl_ptr = 0;
		return;
	}

	nladdr.nl_family = AF_NETLINK;
	nladdr.nl_pid = 0;
	nladdr.nl_groups = 0;

	iov.iov_len = rtnl_ptr;

	status = sendmsg(tp->task_socket, &msg, MSG_DONTWAIT);
	if (status == rtnl_ptr) {
		rtnl_ptr = 0;
	} else {
		trace_log_tp(tp,
			     TRC_NL_BEFORE|TRC_NL_AFTER,
			     LOG_ERR,
			     ("krt_flush: status=%d errno=%d", status, errno));
	}
}

static int
krt_request __PF6(tp, task *,
		  type, u_int,
		  dest, sockaddr_un *,
		  masklen, byte,
		  krtp, krt_parms *,
		  router, sockaddr_un *)
{
	struct nlmsghdr *nlh;
	struct rtmsg *rtm;
	int optr;

	if (rtnl_ptr > 3000)
		rtnl_flush_buffer(tp);
	if (rtnl_ptr > 3000)
		return -ENOBUFS;

	rtnl_ptr = NLMSG_ALIGN(rtnl_ptr);
	nlh = (struct nlmsghdr*)(rtnl_buffer+rtnl_ptr);
	rtm = NLMSG_DATA(nlh);

	nlh->nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	nlh->nlmsg_type = type;
	nlh->nlmsg_flags = NLM_F_CREATE|NLM_F_REPLACE|NLM_F_REQUEST;
	nlh->nlmsg_pid = rtnl_addr.nl_pid;
	nlh->nlmsg_seq = ++rtnl_seq;

	memset(rtm, 0, sizeof(*rtm));
	rtm->rtm_family = AF_INET;
	rtm->rtm_type = RTN_UNICAST;
	rtm->rtm_scope = RT_SCOPE_UNIVERSE;

	if (!BIT_TEST(krtp->krtp_state, RTS_GATEWAY)) {
		rtm->rtm_scope = RT_SCOPE_LINK;
	}
	if (BIT_TEST(krtp->krtp_state, RTS_REJECT)) {
		rtm->rtm_type = RTN_UNREACHABLE;
		rtm->rtm_scope = RT_SCOPE_UNIVERSE;
	}
	if (BIT_TEST(krtp->krtp_state, RTS_BLACKHOLE)) {
		rtm->rtm_type = RTN_BLACKHOLE;
		rtm->rtm_scope = RT_SCOPE_UNIVERSE;
	}
	if (type == RTM_DELETE)
		rtm->rtm_scope = RT_SCOPE_NOWHERE;

	if (krtp->krtp_protocol == RTPROTO_KERNEL) {
		rtm->rtm_protocol = RTPROT_BOOT;
		if (type == RTM_DELETE)
			rtm->rtm_protocol = 0;
		else if (BIT_TEST(krtp->krtp_state, RTS_STATIC))
			rtm->rtm_protocol = RTPROT_STATIC;
	} else {
		if (type == RTM_DELETE) {
			rtm->rtm_protocol = 0;
		}
		else if (krtp->krtp_protocol == RTPROTO_AGGREGATE)
			rtm->rtm_protocol = 254;
		else if (krtp->krtp_protocol == RTPROTO_BGP)
			rtm->rtm_protocol = 253;
		else if (krtp->krtp_protocol == RTPROTO_OSPF)
			rtm->rtm_protocol = 252;
		else if (krtp->krtp_protocol == RTPROTO_OSPF_ASE)
			rtm->rtm_protocol = 251;
		else if (krtp->krtp_protocol == RTPROTO_RIP)
			rtm->rtm_protocol = 250;
		else if (krtp->krtp_protocol == RTPROTO_STATIC)
			rtm->rtm_protocol = 249;
		else if (krtp->krtp_protocol == RTPROTO_DIRECT)
			rtm->rtm_protocol = 248;
		else if (krtp->krtp_protocol == RTPROTO_INET)
			rtm->rtm_protocol = 247;
		else if (krtp->krtp_protocol == RTPROTO_DEFAULT)
			rtm->rtm_protocol = 246;
		else
			rtm->rtm_protocol = RTPROT_GATED;
	}

	rtm->rtm_dst_len = masklen;

	optr = nlh->nlmsg_len;

	RTA_PUT32(nlh, RTA_DST, sock2ip(dest));

	if (rtm->rtm_type == RTN_UNICAST) {
		if (krtp->krtp_n_gw
		    && krtp->krtp_ifaps
		    && krtp->krtp_ifap) {
			RTA_PUT32(nlh, RTA_OIF, krtp->krtp_ifap->ifa_info.ifi_link->ifl_index);
			if (sockaddrcmp(krtp->krtp_ifap->ifa_addr_local, router) ||
			    (sockaddrcmp(krtp->krtp_ifap->ifa_addr, dest) &&
			     krtp->krtp_ifap->ifa_info.ifi_mask_log == masklen))
				rtm->rtm_scope = RT_SCOPE_LINK;
			if (krtp->krtp_ifap->ifa_state&IFS_LOOPBACK)
				rtm->rtm_scope = RT_SCOPE_HOST;
		}
		if (router == inet_addr_loopback)
			rtm->rtm_scope = RT_SCOPE_HOST;
		if (rtm->rtm_scope < RT_SCOPE_LINK)
			RTA_PUT32(nlh, RTA_GATEWAY, sock2ip(router));
		else if (rtm->rtm_scope < RT_SCOPE_HOST && type != RTM_DELETE)
			RTA_PUT32(nlh, RTA_PREFSRC, sock2ip(router));
	}

	if (class_1) {
		rtm->rtm_table = class_1;
	} else {
#if 0
		if (masklen == 0)
			rtm->rtm_table = RT_TABLE_DEFAULT;
		else
#endif
			rtm->rtm_table = RT_TABLE_MAIN;
	}
#ifdef KRT_REALMS
	if (type != RTM_DELETE || krtp->krtp_realm) {
		__u32 realm = krtp->krtp_realm;
		RTA_PUT32(nlh, RTA_FLOW, realm);
	}
#endif

	rtnl_ptr += nlh->nlmsg_len;

	return 0;
}


int krt_change_start(task * tp)
{
	return KRT_OP_SUCCESS;
}


int krt_change_end(task * tp)
{
	if (rtnl_ptr)
		rtnl_flush_buffer(tp);
	return KRT_OP_SUCCESS;
}

int
krt_change __PF5(tp, task *,
		 dest, sockaddr_un *,
		 masklen, byte,
		 old_krtp, krt_parms *,
		 new_krtp, krt_parms *)
{
	int pri = 0;
	int rc = KRT_OP_SUCCESS;
	sockaddr_un *new_router = (sockaddr_un *) 0;
	sockaddr_un *old_router = (sockaddr_un *) 0;

	if (new_krtp) {
		if (!new_krtp->krtp_n_gw) {
			new_router = krt_make_router(socktype(dest), new_krtp->krtp_state);
			assert(new_router);
		} else {
			new_router = new_krtp->krtp_router;
		}
	}

	if (old_krtp) {
		if (!old_krtp->krtp_n_gw) {
			old_router = krt_make_router(socktype(dest), old_krtp->krtp_state);
			assert(old_router);
		} else {
			old_router = old_krtp->krtp_router;
		}
	}

	if (new_krtp) {

		if (old_krtp
		    && old_krtp->krtp_state == new_krtp->krtp_state
		    && (!old_krtp->krtp_n_gw || old_krtp->krtp_ifap)
		    &&
		    ( BIT_TEST(new_krtp->krtp_state, RTS_REJECT|RTS_BLACKHOLE)
		     || (sockaddrcmp(old_router, new_router)
			 && new_krtp->krtp_ifap == old_krtp->krtp_ifap
#ifdef KRT_REALMS
			 && new_krtp->krtp_realm == old_krtp->krtp_realm
#endif
			 && new_krtp->krtp_protocol == old_krtp->krtp_protocol
			 && (new_krtp->krtp_protocol != RTPROTO_KERNEL ||
			     !((new_krtp->krtp_state^old_krtp->krtp_state)&RTS_STATIC)
			     )))) {
			/* If nothing has changed, there isn't anything to do */
			return rc;
		}

		switch (krt_request(tp, RTM_ADD, dest, masklen, new_krtp, new_router)) {
		case ENOBUFS:
			/* Out of resources, wait until later */
			return KRT_OP_FULL;

		case ENETUNREACH:
		case ENETDOWN:
		case ENODEV:
		case EADDRNOTAVAIL:
			/* Interface down?  Wait a while and it might get fixed */
			return KRT_OP_DEFER;

		default:
			krt_trace(tp,
				  "SEND",
				  "CHANGE",
				  dest,
				  masklen,
				  new_router,
				  (flag_t) krt_state_to_flags(new_krtp->krtp_state),
#ifdef KRT_REALMS
				  new_krtp->krtp_realm,
#endif
				  strerror(errno),
				  LOG_CRIT);
			return KRT_OP_NOCANDO;

		case 0:
			krt_trace(tp,
				  "SEND",
				  "CHANGE",
				  dest,
				  masklen,
				  new_router,
				  (flag_t) krt_state_to_flags(new_krtp->krtp_state),
#ifdef KRT_REALMS
				  new_krtp->krtp_realm,
#endif
				  NULL,
				  0);
			return rc;
		}
	}

	if (old_krtp) {
		switch (krt_request(tp, RTM_DELETE, dest, masklen, old_krtp, old_router)) {
		case ENOBUFS:
			/* Not enough resources to perform the action */
			return KRT_OP_FULL;

		default:
			/* Should not happen */
			krt_trace(tp,
				  "SEND",
				  "DELETE",
				  dest,
				  masklen,
				  old_router,
				  (flag_t) krt_state_to_flags(old_krtp->krtp_state),
#ifdef KRT_REALMS
				  old_krtp->krtp_realm,
#endif
				  strerror(errno),
				  LOG_CRIT);
			return KRT_OP_NOCANDO;
		case ESRCH:
			pri = LOG_INFO;
			/* Route not found.  Not a problem since we were trying to delete it anyway */
		case 0:
			if (TRACE_TP(tp, TR_KRT_REQUEST)) {
				krt_trace(tp,
					  "SEND",
					  "DELETE",
					  dest,
					  masklen,
					  old_router,
					  (flag_t) krt_state_to_flags(old_krtp->krtp_state),
#ifdef KRT_REALMS
					  old_krtp->krtp_realm,
#endif
					  pri ? (const char *) strerror(errno) : (char *) 0,
					  pri);
				pri = 0;
			}
			
			return KRT_OP_SUCCESS;
		}
	}
	return rc;
}

void
krt_delete_dst __PF8(tp, task *,
		     dest, sockaddr_un *,
		     masklen, byte,
		     proto, proto_t,
		     state, flag_t,
		     n_gw, int,
		     routers, sockaddr_un **,
		     ifaps, if_addr **)
{
	krt_parms krtp;

	krtp.krtp_protocol = proto;
	krtp.krtp_state = state | RTS_GATEWAY;
	krtp.krtp_n_gw = n_gw;
	krtp.krtp_routers = routers;
	krtp.krtp_ifaps = ifaps;

	(void) krt_change(tp,
			  dest,
			  masklen,
			  &krtp,
			  (krt_parms *) 0);
}

void __krt_recv_ll(task * tp, struct nlmsghdr *nlh, int err)
{
	struct ifinfomsg *ifi = NLMSG_DATA(nlh);
	char * ifname;
	unsigned mtu;
	byte * lladdr = "nil";
	int lla_len = 0;

	struct rtattr *rta[IFLA_MAX+1];

	memset(rta, 0, sizeof(rta));
	parse_rtattr(rta, IFLA_MAX, IFLA_RTA(ifi), IFLA_PAYLOAD(nlh));
	ifname = rta[IFLA_IFNAME] ? RTA_DATA(rta[IFLA_IFNAME]) : "nil";
	mtu = rta[IFLA_MTU] ? *(unsigned*)RTA_DATA(rta[IFLA_MTU]) : 0;
	if (rta[IFLA_ADDRESS]) {
		lladdr = RTA_DATA(rta[IFLA_ADDRESS]);
		lla_len = RTA_PAYLOAD(rta[IFLA_ADDRESS]);
	}

	trace_only_tp(tp,
		      0,
		      ("krt_recv_ll: %d %s", ifi->ifi_index, ifname));

	ifl_addup(tp,
		  ifl_locate_index(ifi->ifi_index),
		  ifi->ifi_index,
		  krt_if_flags(ifi->ifi_flags),
		  0,
		  (mtu_t) mtu,
		  ifname,
		  strlen(ifname),
		  sockbuild_ll(LL_8022, lladdr, lla_len));
}

void krt_recv_ll(task * tp, struct nlmsghdr *nlh)
{
	if_conf_open(tp, FALSE);
	__krt_recv_ll(tp, nlh, 0);
	if_conf_close(tp, TRUE);
}

void krt_recv_ifaddr(task * tp, struct nlmsghdr *nlh)
{
	if_conf_open(tp, FALSE);
	__krt_recv_ifaddr(tp, nlh, 0);
	if_conf_close(tp, FALSE);
}

void krt_recv_err(task * tp, struct nlmsghdr *nlh)
{
	struct nlmsgerr *nle = NLMSG_DATA(nlh);
	struct nlmsghdr *eh;
	
	if (nlh->nlmsg_pid != rtnl_addr.nl_pid)
		return;
	if (nle->error == 0)
		return;

	eh = &nle->msg;

	trace_log_tp(tp,
		     TRC_NL_BEFORE|TRC_NL_AFTER,
		     LOG_ERR,
		     ("krt_recv_err: errno=%d type=%d", nle->error, eh->nlmsg_type));

	if (eh->nlmsg_type != RTM_NEWROUTE)
		return;

	if (nle->error != -ENOBUFS && nle->error != -EADDRNOTAVAIL
	    && nle->error != -ENODEV && nle->error != -ENETUNREACH)
		return;

	eh->nlmsg_type = RTM_DELROUTE;
	eh->nlmsg_pid = 0;
	eh->nlmsg_flags &= ~NLM_F_REQUEST;
	if (NLMSG_LENGTH(4+eh->nlmsg_len) < nlh->nlmsg_len)
		return;
	__krt_recv_route(tp, eh, -nle->error);
}

static char rtnl_recv_buf[8192];

int rtnl_receive_packet __PF2(tp, task *,
			      count, size_t *)
{
	int    rc;
	static char cbuf[64];
	struct sockaddr_nl peer;
	static struct iovec iov = { rtnl_recv_buf, sizeof(rtnl_recv_buf) };
	struct msghdr msg = {
		(void*)&peer, sizeof(peer),
		&iov,	1,
		&cbuf,	sizeof(cbuf),
		0 };

	peer_creds = NULL;
	rc = recvmsg(tp->task_socket, &msg, MSG_DONTWAIT);

	if (rc >= 0) {
		struct cmsghdr *cmsg;

#ifdef KRT_NETLINK2_DEBUG
		trace_log_tp(tp,
			     TRC_NL_BEFORE|TRC_NL_AFTER,
			     LOG_ERR,
			     ("rtnl_recv: message from %d to %08x",
			      peer.nl_pid, peer.nl_groups));
#endif

		if (msg.msg_flags&MSG_TRUNC) {
			trace_log_tp(tp,
				     TRC_NL_BEFORE|TRC_NL_AFTER,
				     LOG_ERR,
				     ("rtnl_recv: message is truncated"));
			return -EINVAL;
		}
		if (msg.msg_flags&MSG_CTRUNC) {
			trace_log_tp(tp,
				     TRC_NL_BEFORE|TRC_NL_AFTER,
				     LOG_ERR,
				     ("rtnl_recv: control message is truncated"));
			return -EINVAL;
		}
		if (msg.msg_namelen != sizeof(peer) ||
		    peer.nl_family != AF_NETLINK ||
		    peer.nl_pid) {
			trace_log_tp(tp,
				     TRC_NL_BEFORE|TRC_NL_AFTER,
				     LOG_ERR,
				     ("rtnl_recv: invalid sender"));
			return -EINVAL;
		}
		for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
			if (cmsg->cmsg_level != SOL_SOCKET ||
			    cmsg->cmsg_type != SCM_CREDENTIALS ||
			    cmsg->cmsg_len < CMSG_LEN(sizeof(struct ucred))) {
				trace_log_tp(tp,
					     TRC_NL_BEFORE|TRC_NL_AFTER,
					     LOG_ERR,
					     ("rtnl_recv: unexpected control message"));
				return -EINVAL;
			}
			peer_creds = (struct ucred*)CMSG_DATA(cmsg);
		}

		if (peer_creds == NULL || peer_creds->uid != 0) {
			trace_log_tp(tp,
				     TRC_NL_BEFORE|TRC_NL_AFTER,
				     LOG_ERR,
				     ("rtnl_recv: access violation: uid=%d", peer_creds ? peer_creds->uid : -1));
			return -EINVAL;
		}

		*count = rc;
		return 0;
	}
	return errno;
}

void krt_recv(task * tp)
{
	int n_packets = TASK_PACKET_LIMIT;
	size_t size;

	while (n_packets--) {
		struct nlmsghdr *nlh;
		int err = rtnl_receive_packet(tp, &size);

		if (err) {
			switch (err) {
			case ENOBUFS:
				/* Overrun! */
				/* Flush our receive queue */
				for (;;) {
					err = rtnl_receive_packet(tp, &size);
					if (err != 0 && err != ENOBUFS)
						break;
				}
				/* We are out of sync.... */
				krt_ifread(task_state);
				return;
			default:
				return;
			}
		}

		nlh = (struct nlmsghdr *)rtnl_recv_buf;

		while (NLMSG_OK(nlh, size)) {

			if (nlh->nlmsg_flags&NLM_F_REQUEST) {
				nlh = NLMSG_NEXT(nlh, size);
				continue;
			}

			switch (nlh->nlmsg_type) {
			case RTM_NEWROUTE:
			case RTM_DELROUTE:
				if (nlh->nlmsg_len >= NLMSG_LENGTH(sizeof(struct rtmsg)))
					__krt_recv_route(tp, nlh, 0);
				break;

			case RTM_NEWADDR:
			case RTM_DELADDR:
				if (nlh->nlmsg_len >= NLMSG_LENGTH(sizeof(struct ifaddrmsg)))
					krt_recv_ifaddr(tp, nlh);
				break;

			case RTM_NEWLINK:
			case RTM_DELLINK:
				if (nlh->nlmsg_len >= NLMSG_LENGTH(sizeof(struct ifinfomsg)))
					krt_recv_ll(tp, nlh);
				break;

			case NLMSG_ERROR:
				if (nlh->nlmsg_len >= NLMSG_LENGTH(sizeof(struct nlmsgerr)))
					krt_recv_err(tp, nlh);
				break;

			case NLMSG_DONE:

			default:
				/* Do nothing: unknown message type */
				break;
			}

			nlh = NLMSG_NEXT(nlh, size);
		}
	}
}


__u32 rtnl_list_seq;
struct sockaddr_nl rtnl_list_addr;
int rtnl_list_type;

static int krt_recv_list(task *tp, int fd, void (*rcv)(task *, struct nlmsghdr*, int))
{
	char	buf[8192];
	char	cbuf[64];
	struct sockaddr_nl peer;
	struct iovec iov = { buf, sizeof(buf) };
	struct ucred *uc;

	while (1) {
		int status;
		struct nlmsghdr *h;
		struct cmsghdr *cmsg;

		struct msghdr msg = {
			(void*)&peer, sizeof(peer),
			&iov,	1,
			&cbuf,	sizeof(cbuf),
			0
		};

		status = recvmsg(fd, &msg, 0);

		if (status < 0) {
			if (errno == EINTR)
				continue;
			/* Though... also continue :-) */
			continue;
		}
		if (status == 0)
			return -1;

		if (msg.msg_namelen != sizeof(peer))
			return -1;
		if (msg.msg_flags & (MSG_TRUNC|MSG_CTRUNC))
			return -1;

		uc = NULL;
		for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
			if (cmsg->cmsg_level != SOL_SOCKET ||
			    cmsg->cmsg_type != SCM_CREDENTIALS ||
			    cmsg->cmsg_len < CMSG_LEN(sizeof(struct ucred))) {
				trace_log_tp(tp,
					     TRC_NL_BEFORE|TRC_NL_AFTER,
					     LOG_ERR,
					     ("rtnl_recv: unexpected control message"));
				return -EINVAL;
			}
			uc = (struct ucred*)CMSG_DATA(cmsg);
		}

		if (uc == NULL || uc->uid != 0) {
			trace_log_tp(tp,
				     TRC_NL_BEFORE|TRC_NL_AFTER,
				     LOG_ERR,
				     ("rtnl_recv_list: access violation: uid=%d", uc ? uc->uid : -1));
			return -EINVAL;
		}

		h = (struct nlmsghdr*)buf;
		while (NLMSG_OK(h,status)) {

			if (h->nlmsg_pid != rtnl_list_addr.nl_pid ||
			    h->nlmsg_seq != rtnl_list_seq)
				goto skip_it;

			if (h->nlmsg_type == NLMSG_DONE)
				return 0;

			if (h->nlmsg_type == NLMSG_ERROR) {
				struct nlmsgerr *err = (struct nlmsgerr*)NLMSG_DATA(h);
				if (h->nlmsg_len >= NLMSG_LENGTH(sizeof(struct nlmsgerr)))
					errno = -err->error;
				return -1;
			}

			if (h->nlmsg_type == rtnl_list_type)
				rcv(tp, h, 0);

skip_it:
			h = NLMSG_NEXT(h, status);
		}
	}
}

static int rtnl_tune(int fd, struct sockaddr_nl *nladdr, unsigned subscriptions)
{
	int addr_len;
	int on = 1;

	memset(nladdr, 0, sizeof(sizeof nladdr));
	nladdr->nl_family = AF_NETLINK;
	nladdr->nl_groups = subscriptions;

	if (bind(fd, (struct sockaddr*)nladdr, sizeof(*nladdr)) < 0) {
		close(fd);
		return -1;
	}
	addr_len = sizeof(*nladdr);
	if (getsockname(fd, (struct sockaddr*)nladdr, &addr_len) < 0) {
		close(fd);
		return -1;
	}
	if (addr_len != sizeof(*nladdr)) {
		close(fd);
		return -1;
	}
	if (nladdr->nl_family != AF_NETLINK) {
		close(fd);
		return -1;
	}
	setsockopt(fd, SOL_SOCKET, SO_PASSCRED, (void*)&on, sizeof(on));
	return fd;
}

static int rtnl_open(struct sockaddr_nl *nladdr)
{
	int fd;

	fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
	if (fd < 0)
		return -1;

	if (rtnl_tune(fd, nladdr, 0) < 0) {
		close(fd);
		return -1;
	}

	return fd;
}

static int rtnl_send_dump_req(int fd, struct sockaddr_nl *nladdr, int type)
{
	int err;
	struct sockaddr_nl knladdr;
	struct {
		struct nlmsghdr nlh;
		struct rtgenmsg g;
	} req;

	memset(&req, 0, sizeof(req));
	memset(&knladdr, 0, sizeof(knladdr));
	knladdr.nl_family = AF_NETLINK;

	req.g.rtgen_family = AF_INET;
	req.nlh.nlmsg_type = type;
	req.nlh.nlmsg_flags = NLM_F_ROOT|NLM_F_REQUEST;
	req.nlh.nlmsg_len = sizeof(req);
	req.nlh.nlmsg_seq = rtnl_list_seq = ++rtnl_seq;
	memcpy(&rtnl_list_addr, nladdr, sizeof(*nladdr));
	rtnl_list_type = type;

	err = sendto(fd, (void*)&req, sizeof(req), 0, (struct sockaddr*)&knladdr, sizeof(knladdr));
	if (err == sizeof(req))
		return 0;
	return -1;
}

int krt_ifread (flag_t save_task_state)
{
	int fd;
	register task *tp = krt_task;
	struct sockaddr_nl nladdr;

	trace_only_tp(tp,
		      TRC_NL_BEFORE,
		      ("krt_ifread: read iflist (via rtnetlink):"));

	fd = rtnl_open(&nladdr);
	if (fd < 0)
		task_quit(EINVAL);

	if (rtnl_send_dump_req(fd, &nladdr, RTM_GETLINK) < 0) {
		close(fd);
		task_quit(EINVAL);
	}

	if_conf_open(tp, TRUE);

	rtnl_list_type = RTM_NEWLINK;
	krt_recv_list(tp, fd, __krt_recv_ll);

	if (rtnl_send_dump_req(fd, &nladdr, RTM_GETADDR) < 0) {
		close(fd);
		task_quit(EINVAL);
	}

	rtnl_list_type = RTM_NEWADDR;
	if (krt_recv_list(tp, fd, __krt_recv_ifaddr) < 0) {
		close(fd);
		task_quit(EINVAL);
	}

	if_conf_close(tp, FALSE);
	close(fd);
	return 0;
}

int krt_rtread(task * tp)
{
	int fd;
	struct sockaddr_nl nladdr;

	trace_only_tp(tp,
		      TRC_NL_BEFORE,
		      ("krt_rtread: Initial routes read from kernel (via rtnetlink):"));

	fd = rtnl_open(&nladdr);
	if (fd < 0)
		task_quit(EINVAL);

	if (rtnl_send_dump_req(fd, &nladdr, RTM_GETROUTE) < 0) {
		close(fd);
		task_quit(EINVAL);
	}

	rtnl_rtread_active++;
	rtnl_list_type = RTM_NEWROUTE;
	krt_recv_list(tp, fd, __krt_recv_route);
	rtnl_rtread_active--;

	close(fd);

	rtnl_tune(tp->task_socket, &rtnl_addr,
		  RTMGRP_IPV4_ROUTE|RTMGRP_IPV4_IFADDR|RTMGRP_LINK);

	return 0;
}
