/*
 * rt_hash.c
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>
 *
 */

/*
   Dumb hash table replacing radix tree. It is simpler, faster on
   routing table of moderate size and eats much less memory for
   large tables.
 */

#define MALLOC_OK
#define	INCLUDE_RT_VAR

#include "include.h"
#include "inet.h"

struct fib_zone
{
	struct fib_zone	*fz_next;
	rt_head		**fz_hash;
	int		fz_nent;
	int		fz_divisor;
	unsigned	fz_hashmask;
	int		fz_logmask;
	unsigned	fz_mask;
	unsigned	fz_netmask;
};

typedef struct _rt_table
{
    struct _rt_table	*rt_table_next;	/* next table in list */
    struct fib_zone	*fib_zones[33];
    struct fib_zone	*fib_zone_list;
    struct sock_info *rt_table_sinfo;	/* stuff we need to grok sockaddr's */
    u_long		rt_table_routes;	/* number of routes in tree */
} rt_table;

static block_t rt_table_block_index;
static block_t fib_zone_block_index;

static rt_table *rt_tables[AF_MAX] = {0};
static rt_table *rt_table_list = (rt_table *) 0;

/* Some method for manipulating the rt_table structure
 */
u_long
rt_table_nodes __PF1(af, int)
{
	return(rt_tables[af]? rt_tables[af]->rt_table_routes: 0L);
}

u_long
rt_table_routes __PF1(af, int)
{
        return(rt_tables[af]? rt_tables[af]->rt_table_routes: 0L);
}

#define	RT_TABLE_PTR(af)	rt_tables[af]

#define	RT_TABLES() \
    do { \
	register rt_table *class; \
	for (class = rt_table_list; \
	     class; \
	     class = class->rt_table_next) { \

#define	RT_TABLES_END()    } } while (0)


#define FZ_MAX_DIVISOR 1024

static __inline__ unsigned fib_hash(unsigned key, unsigned mask)
{
	unsigned h;
	h = key^(key>>20);
	h = h^(h>>10);
	h = h^(h>>5);
	return h & mask;
}

static __inline__ rt_head ** fz_hash_p(unsigned key, struct fib_zone *fz)
{
	return &fz->fz_hash[fib_hash(key, fz->fz_hashmask)];
}

static __inline__ rt_head * fz_hash(unsigned key, struct fib_zone *fz)
{
	return fz->fz_hash[fib_hash(key, fz->fz_hashmask)];
}

static __inline__ unsigned fib_mask(int logmask)
{
	if (logmask >= 32)
		return 0;
	return htonl(~((1<<logmask)-1));
}

#define INMATCH(fz,f,daddr) ((((daddr)^(f)->rth_dest->in.gin_addr.s_addr)&(fz)->fz_netmask) == 0)


rt_head *
rt_table_locate (sockaddr_un * addr, byte bitlen)
{
	struct fib_zone * fz;
	register rt_table *rtt;
	rt_head *f;
	unsigned daddr = addr->in.gin_addr.s_addr;
	unsigned dst = ntohl(daddr);
	unsigned key;

	if (socktype(addr) != AF_INET || !(rtt = RT_TABLE_PTR(socktype(addr))))
		return NULL;

	assert(bitlen<=32);

	fz = rtt->fib_zones[32-bitlen];
	if (fz == NULL)
		return NULL;

	key = (dst&fz->fz_mask)>>fz->fz_logmask;

	for (f = fz_hash(key, fz); f; f = f->fib_next) {
		if (f->rth_dest->in.gin_addr.s_addr == daddr)
			return f;
	}
	return NULL;
}

/*
 *	Return a list of potential routes to this destination, most specific first
 */
rt_list *
rthlist_match(sockaddr_un * addr)
{
	rt_list *rtl = (rt_list *) 0;
	struct fib_zone * fz;
	register rt_table *rtt;
	rt_head *f;
	unsigned daddr = addr->in.gin_addr.s_addr;
	unsigned dst = ntohl(daddr);

	if (socktype(addr) != AF_INET || !(rtt = RT_TABLE_PTR(socktype(addr))))
		return NULL;

	for (fz = rtt->fib_zone_list; fz; fz = fz->fz_next) {
		unsigned key = (dst&fz->fz_mask)>>fz->fz_logmask;

		for (f = fz_hash(key, fz); f; f = f->fib_next) {
			if (INMATCH(fz,f,daddr)) {
				RTLIST_ADD(rtl, f);
				break;
			}
		}
	}
	return rtl;
}

static __inline__ void fib_rebuild_zone(struct fib_zone *fz,
					rt_head **old_ht,
					int old_divisor)
{
	int i;
	rt_head **ht = fz->fz_hash;
	unsigned hashmask = fz->fz_hashmask;
	rt_head *f, **fp, *next;
	unsigned hash;

	for (i=0; i<old_divisor; i++) {
		for (f=old_ht[i]; f; f=next) {
			unsigned key = ntohl(f->rth_dest->in.gin_addr.s_addr);
			next = f->fib_next;
			f->fib_next = NULL;
			key = (key&fz->fz_mask)>>fz->fz_logmask;
			hash = fib_hash(key, hashmask);
			for (fp = &ht[hash]; *fp; fp = &(*fp)->fib_next)
				/* NONE */;
			*fp = f;
		}
	}
}


static void fib_rehash_zone(struct fib_zone *fz)
{
	rt_head **ht, **old_ht;
	int old_divisor, new_divisor=0;
	unsigned new_hashmask=0;
		
	old_divisor = fz->fz_divisor;

	switch (old_divisor) {
	case 16:
		new_divisor = 256;
		new_hashmask = 0xFF;
		break;
	case 256:
		new_divisor = 1024;
		new_hashmask = 0x3FF;
		break;
	default:
		task_quit(EINVAL);
	}

	ht = task_mem_malloc(NULL, new_divisor*sizeof(rt_head*));

	if (ht)	{
		memset(ht, 0, new_divisor*sizeof(rt_head*));
		old_ht = fz->fz_hash;
		fz->fz_hash = ht;
		fz->fz_hashmask = new_hashmask;
		fz->fz_divisor = new_divisor;
		fib_rebuild_zone(fz, old_ht, old_divisor);
		task_mem_free(NULL, old_ht);
	}
}

static struct fib_zone *
fib_new_zone(rt_table *class, int logmask)
{
	int i;
	struct fib_zone *fz = task_block_alloc(fib_zone_block_index);
	if (!fz)
		return NULL;

	if (logmask < 32) {
		fz->fz_divisor = 16;
		fz->fz_hashmask = 0xF;
	} else {
		fz->fz_divisor = 1;
		fz->fz_hashmask = 0;
	}
	fz->fz_hash = task_mem_malloc(NULL, fz->fz_divisor*sizeof(rt_head*));
	if (!fz->fz_hash) {
		task_block_free(fib_zone_block_index, fz);
		return NULL;
	}
	memset(fz->fz_hash, 0, fz->fz_divisor*sizeof(rt_head*));
	fz->fz_logmask = logmask;
	fz->fz_netmask = fib_mask(logmask);
	fz->fz_mask = ntohl(fz->fz_netmask);
	for (i=logmask-1; i>=0; i--)
		if (class->fib_zones[i])
			break;
	if (i<0) {
		fz->fz_next = class->fib_zone_list;
		class->fib_zone_list = fz;
	} else {
		fz->fz_next = class->fib_zones[i]->fz_next;
		class->fib_zones[i]->fz_next = fz;
	}
	class->fib_zones[logmask] = fz;
	return fz;
}

#define FIB_SCAN(f, fp) \
for ( ; ((f) = *(fp)) != NULL; (fp) = &(f)->fib_next)

#define FIB_CONTINUE(f, fp) \
{ \
	fp = &f->fib_next; \
	continue; \
}


/*
 *	Insert this head into the tree
 */
void
rt_table_add(rt_head * f)
{
	rt_table *class;
	struct fib_zone * fz;

	int logmask = 32 - (int)(f->rth_dest_mask_log);
	unsigned daddr = f->rth_dest->in.gin_addr.s_addr;
	unsigned dst = ntohl(daddr);
	unsigned key;
	rt_head **fp, *f1;

	class = RT_TABLE_PTR(socktype(f->rth_dest));
	assert(class);

	key = dst>>logmask;
	fz = class->fib_zones[logmask];

	if (!fz && !(fz = fib_new_zone(class, logmask)))
		task_quit(ENOMEM);

	if (fz->fz_nent > (fz->fz_divisor<<2) &&
	    fz->fz_divisor < FZ_MAX_DIVISOR &&
	    (!logmask || (1<<(32-logmask)) > fz->fz_divisor))
		fib_rehash_zone(fz);

	fp = fz_hash_p(key, fz);

	/*
	 * Scan list to find a route with the same destination
	 */
	FIB_SCAN(f1, fp) {
		if (f1->rth_dest->in.gin_addr.s_addr == daddr)
			task_quit(EINVAL);
	}

	f->fib_next = f1;
	*fp = f;
	fz->fz_nent++;
	class->rt_table_routes++;
}


/*
 *	Remove this head from the tree
 */
void
rt_table_delete(rt_head * rth)
{
	rt_table *class;
	rt_head **fp, *f;
	struct fib_zone *fz;
	int logmask = 32 - (int)rth->rth_dest_mask_log;
	unsigned daddr = rth->rth_dest->in.gin_addr.s_addr;
	unsigned dst = ntohl(daddr);
	unsigned key;

	assert((class = RT_TABLE_PTR(socktype(rth->rth_dest))));

	fz = class->fib_zones[logmask];

	if (fz != NULL)	{
		key = (dst&fz->fz_mask)>>logmask;
		fp = fz_hash_p(key, fz);

		FIB_SCAN(f, fp) {
			if (f == rth) {
				*fp = f->fib_next;
				f->fib_next = NULL;
				fz->fz_nent--;
				class->rt_table_routes--;
				return;
			}
		}

	}
}

block_t rtlist_block_index = 0;

	    
/* Build a change list of all active routes */
/* 	Routes not to be advised and hidden routes are not located */
rt_list *
rthlist_active __PF1(af, int)
{
    rt_entry *rt;
    rt_list *rtl = (rt_list *) 0;
    struct fib_zone *fz;
    rt_head *f;

    if (af && af != AF_INET)
	    return NULL;

    RT_TABLES()
    for (fz = class->fib_zone_list; fz; fz = fz->fz_next) {
	    int i;
	    for (i=fz->fz_divisor-1; i>=0; i--) {
		    for (f=fz->fz_hash[i]; f; f=f->fib_next) {
			    if ((rt = f->rth_active) == NULL)
				    continue;
			    if (BIT_TEST(rt->rt_state, RTS_NOADVISE|RTS_PENDING)) {
				    /* Not to be announced */
				    continue;
			    }
			    if (BIT_TEST(rt->rt_state, RTS_HIDDEN|RTS_DELETE)) {
				    /* Deleted or hidden, unless in holddown */
				    continue;
			    }
			    RTLIST_ADD(rtl, f);
		    }
	    }
    }
    RT_TABLES_END();
    return rtl ? rtl->rtl_root : rtl;
}


rt_list *
rtlist_all __PF1(af, int)
{
    rt_entry *rt;
    rt_list *rtl = (rt_list *) 0;
    struct fib_zone *fz;
    rt_head *f;

    if (af && af != AF_INET)
	    return NULL;

    RT_TABLES()
    for (fz = class->fib_zone_list; fz; fz = fz->fz_next) {
	    int i;
	    for (i=fz->fz_divisor-1; i>=0; i--) {
		    for (f=fz->fz_hash[i]; f; f=f->fib_next) {
			    RT_ALLRT(rt, f) {
				    RTLIST_ADD(rtl, rt);
			    } RT_ALLRT_END(rt, f);
		    }
	    }
    }
    RT_TABLES_END();
    return rtl ? rtl->rtl_root : rtl;
}


rt_list *
rthlist_all __PF1(af, int)
{
    register rt_head *f;
    register rt_list *rtl = (rt_list *) 0;
    struct fib_zone *fz;

    if (af && af != AF_INET)
	    return NULL;

    RT_TABLES()
    for (fz = class->fib_zone_list; fz; fz = fz->fz_next) {
	    int i;
	    for (i=fz->fz_divisor-1; i>=0; i--) {
		    for (f=fz->fz_hash[i]; f; f=f->fib_next) {
			    RTLIST_ADD(rtl, f);
		    }
	    }
    }
    RT_TABLES_END();
    return rtl ? rtl->rtl_root : rtl;
}


/*
 *	Display information about the radix tree, including a visual
 *	representation of the tree itself 
 */
void
rt_table_dump __PF2(tp, task *,
		    fd, FILE *)
{
    register rt_head *f;
    struct fib_zone *fz;

    RT_TABLES();
    fprintf(fd, "\tLogmask trie for %s (%d) routes %u\n",
	    gd_lower(trace_value(task_domain_bits,
				 (int)class->rt_table_sinfo->si_family)),
	    class->rt_table_sinfo->si_family,
	    class->rt_table_routes);
    for (fz = class->fib_zone_list; fz; fz = fz->fz_next) {
	    int i;
	    int tot = fz->fz_nent;
	    fprintf(fd, "%d %d entries\n", fz->fz_logmask, fz->fz_nent);
	    for (i=fz->fz_divisor-1; i>=0; i--) {
		    int cnt = 0;
		    for (f=fz->fz_hash[i]; f; f=f->fib_next)
			    cnt++;
		    tot -= cnt;
		    fprintf(fd, "\t%d -> %d\n", i, cnt);
	    }
	    if (tot)
		    fprintf(fd, "\t!!!! mismatch %d\n", tot);
    }
    RT_TABLES_END();

    fprintf(fd, "\n");
}


void
rt_table_init_family __PF1(af, int)
{
    rt_table *rtt, *rtt_prev, *rtt_next;

    if (af != AF_INET)
	    task_quit(EINVAL);

    trace_only_tf(trace_global,
		  0,
		  ("rt_table_init_family: Initializing radix tree for %s",
		   trace_value(task_domain_bits, af)));

    rtt = (rt_table *) task_block_alloc(rt_table_block_index);
    rtt->rt_table_sinfo = &(sock_info[af]);
    rt_tables[af] = rtt;

    rtt_prev = (rt_table *) 0;
    rtt_next = rt_table_list;
    while (rtt_next && rtt_next->rt_table_sinfo->si_family < af) {
	rtt_prev = rtt_next;
	rtt_next = rtt_prev->rt_table_next;
    }

    rtt->rt_table_next = rtt_next;
    if (rtt_prev) {
	rtt_prev->rt_table_next = rtt;
    } else {
	rt_table_list = rtt;
    }

    rt_tables[af] = rtt;

    trace_only_tf(trace_global,
		  0,
		  (NULL));
}


/*
 *	Initialize rt heads for all protocols 
 */
void
rt_table_init __PF0(void)
{
    rt_table_block_index = task_block_init(sizeof (rt_table), "rt_table");
    fib_zone_block_index = task_block_init(sizeof (struct fib_zone), "fib_zone");
}

#ifdef PROTO_GII

/* VERY UGLY(TM) tree walker. Well, rt_radix in gated-3.5 has not even
   such one :-(
*/

rtwalk_t *
rt_walk_alloc __PF0(void)
{
	rtwalk_t *rtw;

	rtw = malloc(sizeof(*rtw));
	memset(rtw, 0, sizeof(*rtw));
	return(rtw);
}

rtwalk_t *
rt_walk_free __PF1(rtw, rtwalk_t *)
{
	free(rtw);
	return((rtwalk_t *) NULL);
}

rt_head *
rt_walk_down __PF1(rtw, rtwalk_t *)
{
    register rt_head *f;
    struct fib_zone *fz;
    sockaddr_un *mask = inet_static_mask(rtw->rw_len);
    int len;
    int n=0;

    RT_TABLES()
    for (len=rtw->rw_len; len<=32; len++) {
	    int i;
	    fz = class->fib_zones[32-len];
	    if (fz==NULL)
		    continue;
	    for (i=fz->fz_divisor-1; i>=0; i--) {
		    for (f=fz->fz_hash[i]; f; f=f->fib_next) {
			    if ((rtw->d^f->rth_dest->in.gin_addr.s_addr)&mask->in.gin_addr.s_addr)
				    continue;
			    if (n < rtw->n) {
				    n++;
				    continue;
			    }
			    rtw->n = n+1;
			    rtw->rw_rth = f;
			    return f;
		    }
	    }
    }
    RT_TABLES_END();
    return NULL;
}

rt_head *
rt_walk_up __PF1(rtw, rtwalk_t *)
{
    register rt_head *f;
    struct fib_zone *fz;
    int n=0;
    int len;

    RT_TABLES()
    for (len=rtw->rw_len; len>=0; len--) {
	    int i;
	    fz = class->fib_zones[32-len];
	    if (fz==NULL)
		    continue;
	    for (i=fz->fz_divisor-1; i>=0; i--) {
		    for (f=fz->fz_hash[i]; f; f=f->fib_next) {
			    if ((rtw->d^f->rth_dest->in.gin_addr.s_addr)&fz->fz_netmask)
				    continue;
			    if (n < rtw->n) {
				    n++;
				    continue;
			    }
			    rtw->n = n+1;
			    rtw->rw_rth = f;
			    return f;
		    }
	    }
    }
    RT_TABLES_END();
    return NULL;
}

rt_head *
rt_walk __PF1(rtw, rtwalk_t *)
{
    if (rtw->rw_way == RTW_UP)
	return rt_walk_up(rtw);
    else
	return rt_walk_down(rtw);
}
#endif
