/*
 * task_mem.c
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>
 *
 * Portions of this software may fall under the following copyrights:
 *
 * Copyright (c) 1996, 1997 The Regents of the University of Michigan
 * Copyright (c) 1990,1991,1992,1993,1994,1995 by Cornell University.
 * Copyright (c) 1988 Regents of the University of California.
 *
 *  	include "task.c"
 */

/* This allocator is more friendly to system, than native gated's one.
   However it is slower, so that you prefer gated's one if you are
   not too tough on memory.
 */

#include "include.h"
#include <sys/mman.h>

/* This limit on eaten core is too strict for BGP routers, certainly. */
#define MAX_CORE (16*1024*1024)	/* 16Mb */

#define __MAX(a,b) ((a) < (b) ? (b) : (a))
#define TASK_BLOCK_ALIGN	sizeof(unsigned long)
#define TASK_BLOCK_PAGE_MASK	(~(unsigned long)(task_pagesize-1))
#define	TASK_BLOCK_WITHHEAD(size) __MAX(size, sizeof(struct block_header))
#define	TASK_BLOCK_ROUNDUP(size) ROUNDUP(TASK_BLOCK_WITHHEAD(size), TASK_BLOCK_ALIGN)

/*
 * A free block header.
 */

struct block_header
{
	unsigned long		bh_magic;
	struct block_header	*bh_next;
};

struct tuple
{
	struct tuple *next;
	struct tuple *prev;
};

struct page_descriptor
{
	struct tuple		link;
	struct block_header	*firstfree;
	int			nfree;
};

#define MF_FREE 0xDEADDEADUL

#define BH(p) ((struct block_header *)(p))
#define PAGE_DESC(p) ((struct page_descriptor *)(((unsigned long)(p)) & TASK_BLOCK_PAGE_MASK))

static __inline__ void
kmalloc_unlink_page(struct page_descriptor *page)
{
	struct tuple * p=(struct tuple *)page;
	p->prev->next = p->next;
	p->next->prev = p->prev;
}

static __inline__ void
kmalloc_link_page_first(struct tuple *pg, struct page_descriptor *page)
{
	struct tuple *p=(struct tuple *)page;
	p->next = pg->next;
	p->prev = pg;
	p->next->prev = p;
	p->prev->next = p;
}

static __inline__ void
kmalloc_link_page_last(struct tuple *pg, struct page_descriptor *page)
{
	struct tuple *p=(struct tuple *)page;
	p->next = pg;
	p->prev = pg->prev;
	p->next->prev = p;
	p->prev->next = p;
}

struct task_block
{
    struct task_size_block *tb_block;	/* Pointer to size descriptor */
    struct task_block *tb_next;		/* Pointer to next in chain */
    const char *tb_name;    		/* Name for this block */

    u_int	tb_n_init;		/* Number of init requests */
    u_int	tb_n_alloc;		/* Number of alloc requests */
    u_int	tb_n_free;		/* Number of free requests */
};

struct task_size_block
{
    struct task_size_block *tsb_forw;	/* Queue glue */
    struct task_size_block *tsb_back;
    struct task_block tsb_block;	/* First block using this size */

    size_t tsb_size;			/* Size of this block */
    u_int  tsb_count;			/* Number per page */

    u_int  tsb_n_free;
    u_int  tsb_n_pages;
    u_int  tsb_n_full;

    struct tuple tsb_page_list;
};

/* Head of block size queue */
struct task_size_block task_block_head = {
	&task_block_head, &task_block_head
};

/* Pre-allocated block for size blocks */
static struct task_size_block task_block_sizes = {
	&task_block_sizes, &task_block_sizes,
	{
		&task_block_sizes,
		(struct task_block *) 0,
		"task_size_block"
	}
};

static struct task_size_block task_block_names = {
	&task_block_names, &task_block_names,
	{
		&task_block_names,
		(struct task_block *) 0,
		"task_block"
	}
};
static block_t  task_block_name = &task_block_names.tsb_block;

int __tot_mem;

void_t task_block_malloc(size_t size)
{
	size_t          my_size = ROUNDUP(size, task_pagesize);
	void_t		vp;

	vp = (void_t)mmap(0,
		 my_size,
		 PROT_READ | PROT_WRITE | PROT_EXEC,
		 MAP_PRIVATE | MAP_ANONYMOUS,
		 -1,
		 0);
	
	if (vp == GS2A(-1) || (__tot_mem += my_size) > MAX_CORE) {
		trace_log_tf(trace_global,
			     0,
			     LOG_ERR,
			     ("task_block_malloc: alloc(%u), totmem=%u: %m",
			      my_size, __tot_mem));
		task_quit(ENOBUFS);
	}

	return vp;
}

void task_block_reclaim(size_t size, void_t p)
{
	int sz = ROUNDUP(size, task_pagesize);
	munmap((caddr_t)p, sz);
	__tot_mem -= sz;
}

static inline void_t task_page_alloc(void)
{
	return task_block_malloc(task_pagesize);
}

static inline void  task_page_reclaim(void_t p)
{
	task_block_reclaim(task_pagesize, p);
}

void task_block_free(struct task_block *tbp, void *ptr)
{
	struct task_size_block *tsb = tbp->tb_block;
	u_int  n = tsb->tsb_count;
	size_t size = tsb->tsb_size;
	struct block_header *p;
	struct page_descriptor *page;

	tbp->tb_n_free++;

	/* Large block */
	if (n == 1) {
		tsb->tsb_n_pages--;
		task_block_reclaim(size, ptr);
		return;
	}

	p = BH(ptr);
	page = PAGE_DESC(p);

	p->bh_magic = MF_FREE;
	p->bh_next = page->firstfree;
	page->firstfree = p;
	page->nfree++;
	tsb->tsb_n_free++;

	/* If page contains only one free block, it was not
	   on page list, now we put it FIRST.
	 */

	if (page->nfree == 1) {
		tsb->tsb_n_full--;
		kmalloc_link_page_first(&tsb->tsb_page_list, page);
		return;
	}

	kmalloc_unlink_page(page);

	/* If there are something free; put it back on list
	   in correct place.
	 */
	if (page->nfree != n) {
		struct tuple *pg1 = page->link.next;
		struct page_descriptor *page1 = (struct page_descriptor*)pg1;

		while (pg1 != &tsb->tsb_page_list &&
		       page1->nfree > page->nfree) {
			pg1 = pg1->next;
			page1 = (struct page_descriptor*)pg1;
		}
		kmalloc_link_page_last(pg1, page);
		return;
	}

	/* Page is absolutely free; return it to the system */
	task_page_reclaim(page);
	tsb->tsb_n_pages--;
	tsb->tsb_n_free -= tsb->tsb_count;
}

extern void* task_block_alloc(struct task_block *tbp)
{
	struct task_size_block *tsb = tbp->tb_block;
	u_int  n = tsb->tsb_count;
	size_t size = tsb->tsb_size;

	struct block_header *p;
	struct page_descriptor *page;
	struct tuple *pg;
	int i;

	tbp->tb_n_alloc++;

	if (n == 1) {
		tsb->tsb_n_pages++;
		return task_block_malloc(size);
	}

	pg = &tsb->tsb_page_list;
	tsb->tsb_n_free--;

	if (pg->next != pg) {
		page = (struct page_descriptor*)pg->next;
		p = page->firstfree;
		if (p->bh_magic != MF_FREE) {
			syslog(LOG_ERR, "block on freelist at %08lx isn't free.\n", (long) p);
			task_quit(ENOBUFS);
		}
		page->firstfree = p->bh_next;
		--page->nfree;
		memset(p, 0, size);
		if (page->nfree != 0)
			return p;

		/* Page is 100% filled; unlink it from list */
		kmalloc_unlink_page(page);
		tsb->tsb_n_full++;
		return p;
	}

	page = (struct page_descriptor *) task_page_alloc();
	tsb->tsb_n_pages++;
	tsb->tsb_n_free += tsb->tsb_count;

	/* Loop for all but the last block: */
	for (i = n, p = BH(page + 1); i > 1; i--, p = p->bh_next) {
		p->bh_magic = MF_FREE;
		p->bh_next = BH(((long) p) + tsb->tsb_size);
	}
	p->bh_magic = MF_FREE;
	p->bh_next = NULL;
	page->nfree = n - 1;
	p = BH(page+1);
	page->firstfree = p->bh_next;
	p->bh_magic = 0;
	p->bh_next = NULL;
	kmalloc_link_page_last(&tsb->tsb_page_list, page);
	return p;
}

block_t task_block_init (size_t size, const char *name)
{
	register struct task_size_block *tsb = task_block_head.tsb_forw;
	struct task_block *tbp;
	register int    count;

	/* Round up to the minimum size */
	size = TASK_BLOCK_ROUNDUP(size);

	count = (task_pagesize - sizeof(struct page_descriptor))/ size;
	if (count < 1) {
		/* Greater than a page, allocate one at a time */
		count = 1;
	}
	/* Look for this size in current list */
	do {
		if (tsb->tsb_size == size) {
			/* Found the size block */

			/* Check for duplicate request */
			for (tbp = &tsb->tsb_block; tbp; tbp = tbp->tb_next) {
				if (!strcmp(name, tbp->tb_name))
					goto Return;
			}

			/*
			 * Allocate a block for this caller and append it to
			 * the chain
			 */
			tbp = (struct task_block *) task_block_alloc(task_block_name);
			tbp->tb_n_init = 0;
			tbp->tb_next = tsb->tsb_block.tb_next;
			tsb->tsb_block.tb_next = tbp;
			goto Found;
		} else if (size < tsb->tsb_size) {
			/* Insert before this one */

			break;
		}
	} while ((tsb = tsb->tsb_forw) != &task_block_head);

	tsb = tsb->tsb_back;

	/* Allocate a block before this element */
	INSQUE(task_block_alloc(&task_block_sizes.tsb_block), tsb);
	tsb = tsb->tsb_forw;

	/* Fill in the info */
	tsb->tsb_size = size;
	tsb->tsb_count = count;
	tsb->tsb_page_list.next = tsb->tsb_page_list.prev = &tsb->tsb_page_list;
	tbp = &tsb->tsb_block;

Found:
	/* Link the name block back to the size block */
	tbp->tb_block = tsb;

	/* Save the name (must be a constant) */
	tbp->tb_name = name;

Return:
	tbp->tb_n_init++;

	return tbp;
}


void task_block_init_all(void)
{
	/* Init the size block */
	task_block_sizes.tsb_size = TASK_BLOCK_ROUNDUP(sizeof(struct task_size_block));
	task_block_sizes.tsb_count = (task_pagesize - sizeof(struct page_descriptor)) / task_block_sizes.tsb_size;
	task_block_sizes.tsb_page_list.next = 
	task_block_sizes.tsb_page_list.prev = 
	&task_block_sizes.tsb_page_list;
	

	/* Put it on the linked list */
	INSQUE(&task_block_sizes, &task_block_head);

	/* Init the names block */
	task_block_names.tsb_size = TASK_BLOCK_ROUNDUP(sizeof(struct task_block));
	task_block_names.tsb_count = (task_pagesize - sizeof(struct page_descriptor)) / task_block_names.tsb_size;
	task_block_names.tsb_page_list.next = 
	task_block_names.tsb_page_list.prev = 
	&task_block_names.tsb_page_list;

	/* Put it on the linked list */
	INSQUE(&task_block_names, &task_block_head);
}


void task_block_dump(FILE *fp)
{
	register struct task_size_block *tsb = task_block_head.tsb_forw;

	fprintf(fp, "Task Blocks:\n\tAllocation size: %5d\n\n",
		task_pagesize);

	do {
		struct task_block *tbp = &tsb->tsb_block;

		fprintf(fp, "\tSize: %5u\tN_free: %u\tN_chunks: %u\tFull: %u\tRemnants: %u\n",
			tsb->tsb_size,
			tsb->tsb_n_free,
			tsb->tsb_n_pages,
			tsb->tsb_n_full,
			tsb->tsb_count == 1 ? 0 : tsb->tsb_n_pages*(task_pagesize-tsb->tsb_count*tsb->tsb_size));

		do {
			fprintf(fp, "\t%32s\tInit: %8u  Alloc: %8u  Free: %8u  InUse: %8u\n",
				tbp->tb_name,
				tbp->tb_n_init,
				tbp->tb_n_alloc,
				tbp->tb_n_free,
				tbp->tb_n_alloc - tbp->tb_n_free);
		} while ((tbp = tbp->tb_next));

		fprintf(fp, "\n");

	} while ((tsb = tsb->tsb_forw) != &task_block_head);

	fprintf(fp, "\n");
}

